from skilletlib.snippet.workflow import WorkflowSnippet

from sli.decorators import load_variables
from sli.decorators import require_ngfw_connection_params
from sli.decorators import require_single_skillet
from sli.decorators import require_skillet_type
from sli.tools import get_var
from sli.tools import render_expression
from .base import BaseCommand


class WorkflowCommand(BaseCommand):
    sli_command = 'workflow'
    short_desc = 'Execute a workflow skillet'

    @require_single_skillet
    @require_skillet_type('workflow')
    @require_ngfw_connection_params
    @load_variables
    def run(self):

        # Loop over every snippet
        for snippet in self.sli.skillet.snippet_stack:

            # Verify if a snippet should be run
            if not self._should_execute(snippet, self.sli.context):
                continue

            # Handle transform
            if "transform" in snippet:
                workflow_snippet = WorkflowSnippet(snippet, self.sli.skillet, self.sli.sl)
                self.sli.context.update(workflow_snippet.transform_context(self.sli.context))

            # Find and execute specific skillet
            snippet_skillet = [x for x in self.sli.skillets if x.name == snippet['name']]
            if len(snippet_skillet) < 1:
                print(f"Unable to find skillet {snippet['name']}")
                return
            if len(snippet_skillet) > 1:
                print(f"Unable to execute unique skillet {snippet['name']} as {len(snippet_skillet)} were found ")
                return
            snippet_skillet = snippet_skillet[0]
            print(f"Running skillet {snippet_skillet.name} - {snippet_skillet.type}\n")
            for var in snippet_skillet.variables:
                get_var(var, self.args, self.sli.context)
            exe = snippet_skillet.execute(self.sli.context)
            self.sli.context.update(snippet_skillet.context)

            # Handle outputs
            if snippet_skillet.type == 'pan_validation':
                self._print_validation_output(exe)
            elif snippet_skillet.type == 'template':
                print(snippet_skillet.get_results()['template'])
            elif 'output_template' in exe:
                print(exe['output_template'])

    def _get_output(self):
        pass

    @staticmethod
    def _should_execute(snippet, context):
        """Determine if a workflow snippet should be executed"""
        if 'when' not in snippet:
            return True
        return render_expression(snippet['when'], context) == "True"

    @staticmethod
    def _print_validation_output(exe):
        """Print output for validation skillets after run"""
        for output_name in exe["outputs"]:
            output = exe["outputs"][output_name]
            if not isinstance(output, dict):
                continue
            if "test" in output:
                print(f"   Validation: {output_name}")
                print(f"      Output: {output['output_message']}")



from sli.tools import print_table
from .base import BaseCommand

"""
Load and display a list of skillets in a directory.
Display any errors
"""


class LoadCommand(BaseCommand):
    sli_command = 'load'
    short_desc = 'Load and display all skillets of any type'

    skillets = list()

    def run(self):
        """SLI action, test load all skillets in directory and print out loaded skillets"""
        for s in self.sli.skillets:
            obj = {
                'name': s.name,
                'type': s.type,
            }
            self.skillets.append(obj)

    def _get_output(self):
        print_table(
            self.skillets,
            {
                "Name": "name",
                "Type": "type",
            }
        )



import os
from os.path import expanduser
from panforge import Report
from getpass import getpass
import socket
from jinja2 import Environment
import jmespath


def get_var(var, args, context):
    """Get a var by searching args and the context, else ask the user"""
    args = {x.split('=')[0]: x.split('=')[1] for x in args if '=' in x}
    if var['name'] in args:
        # First order of preference is to use a CLI provided parameter
        context[var['name']] = args[var['name']]
    else:
        # If input has not yet been supplied, get it from the user
        context.update(get_variable_input(var, context))


def render_template(template_text, context):
    """Render a template loaded from a string with an appropriate environment"""
    env = Environment(extensions=['jinja2_ansible_filters.AnsibleCoreFiltersExtension'])
    return env.from_string(template_text).render(context)


def render_expression(expression, context):
    """Shortcut to render an expression, adding {{}} brackets to a template"""
    return render_template("{{" + str(expression) + "}}", context)


def is_ip(ip):
    """
    Return True if either an IPv4 or IPV6 address
    """
    for family in socket.AF_INET, socket.AF_INET6:
        try:
            socket.inet_pton(family, ip)
            return True
        except Exception:
            pass
    return False


def print_table(objs, defs):
    """
    Print a formatted table with left justified columns
    """

    def get_cell(text, max):
        if text is None:
            return " " * max
        return f"  {text + ' ' * (max - len(text))}  "

    # Calculate max width of each column
    cols = {k: {"width": len(k)} for k in defs}
    for o in objs:
        for k in defs:
            if "." in defs[k]:
                text = jmespath.search(defs[k], o)
            else:
                text = o[defs[k]]
            if text is None:
                continue
            if len(text) > cols[k]["width"]:
                cols[k]["width"] = len(text)
    # Print headers
    header = "".join([get_cell(x, cols[x]["width"]) for x in cols])
    print(header)
    print("-" * len(header))

    # Print data
    for o in objs:
        row = ""
        for k in defs:
            if "." in defs[k]:
                text = jmespath.search(defs[k], o)
            else:
                text = o[defs[k]]
            row += get_cell(text, cols[k]["width"])
        print(row)


def generate_report(out_file, data, report_dir, header=None):
    """
    Generate a panforge report at 'out_file' from source 'data' using 'report_dir'
    as the root reporting directory for panforge
    """
    if not os.path.exists(report_dir):
        print(f'Could not generate report, source directory {report_dir} not present')
        return
    report = Report(report_dir)
    if header:
        report.load_header(header)
    report.load_data(data)
    report_html = report.render_html()
    with open(out_file, 'w') as f:
        f.write(report_html)
    print(f'Report written to {out_file}')


def input_yes_no(prompt, ret_value, ret_false=None, default=None):
    """
    Simple wrapper for asking user yes no questions until an acceptable answer
    is received, returns ret_value if yes, ret_false if no
    """
    default_str = " (y/n): "
    if isinstance(default, str):
        default_yes = True if default.lower() in ("y", "yes") else False
        default_str = f" (y/n default: {'yes' if default_yes else 'no'}): "
    while True:
        user_input = input(prompt + default_str)
        if not len(user_input) and default is not None:
            return ret_value if default_yes else ret_false
        if user_input.lower() in ('y', 'yes'):
            return ret_value
        if user_input.lower() in ('n', 'no'):
            return ret_false
        print('Please input either y or n')


def check_default_empty(default):
    """Helper function to validate if a default parameter is empty based on type"""
    if isinstance(default, str) or isinstance(default, list):
        return len(default) == 0
    elif isinstance(default, int):
        return False
    elif isinstance(default, dict):
        return len(default.keys()) == 0


def get_variable_input(var, context):
    """
    Get input from user in reference to var, being a skillet variable object
    """
    type_hint = var.get('type_hint')
    name = var.get('name')
    desc = var.get('description', name)

    # First get a default from the context, otherwise use the skillet definition
    default = context.get(name, "")
    default_is_empty = check_default_empty(default)

    if default_is_empty:
        default = var.get("default", "")
        default_is_empty = check_default_empty(default)

    default_str = f"({default})" if not default_is_empty else ""
    if not name:
        raise ValueError('Input variable missing name')

    ret_dict = {}

    # Check for a toggle hint and return if not applicable
    if 'toggle_hint' in var:
        if not var['toggle_hint']['value'] == context.get(var['toggle_hint'].get('source')):
            return ret_dict

    if type_hint == 'checkbox':

        if 'source' in var:
            raw_dd_list = context.get(var["source"], [])
            if not isinstance(raw_dd_list, list):
                raw_dd_list = [raw_dd_list]

            cbx_list = [{"key": x, "value": x} for x in raw_dd_list]

        else:
            cbx_list = var.get('cbx_list', [])

        print(f"\n{desc}\n{'-'*len(desc)}\n")

        # Get inputs from user until user confirms they are ok
        confirmed = False
        input_list = []
        while not confirmed:
            input_list.clear()
            for cbx in cbx_list:
                default_result = None
                if isinstance(default, list):
                    default_result = "y" if cbx["value"] in default else "n"
                input_value = input_yes_no(cbx['key'], cbx['value'], default=default_result)
                if input_value:
                    input_list.append(input_value)

            # Display all inputs
            check = []
            for cbx in cbx_list:
                check.append({'name': cbx['key'], 'value': 'Yes' if cbx['value'] in input_list else 'No'})
            print('')
            print_table(check, {'Input': 'name', 'Value': 'value'})
            print('')
            if input_yes_no('Are These answers ok?', True):
                confirmed = True

        # Add input to ret_dict
        ret_dict[name] = input_list

    elif type_hint == 'dropdown':
        print(f"\n{desc}")

        if 'source' in var:
            raw_dd_list = context.get(var["source"], [])
            if not isinstance(raw_dd_list, list):
                raw_dd_list = [raw_dd_list]

            dd_list = [{"key": x, "value": x} for x in raw_dd_list]
        else:
            dd_list = var.get('dd_list', [])

        for i in range(0, len(dd_list)):
            print(f"   {i + 1}. {dd_list[i]['key']}")

        valid_response = False
        while not valid_response:
            default_key = ""
            if not default_is_empty:
                default_dd = [x for x in dd_list if x['value'] == default]
                if not len(default_dd):
                    # If there is a default from the context and it's not an option, clear the default
                    default = ""
                else:
                    default_dd = default_dd[0]
                    default_key = f"({default_dd['key']})"
            response = input(f"Please enter line number of selection {default_key}: ")
            if not len(response) and not default_is_empty:
                # Assume default
                response = default_dd["value"]
                valid_response = True
            elif response.isdigit() and '.' not in response:
                response_index = int(response) - 1
                if response_index < 0 or response_index >= len(dd_list):
                    print(f"Please input a number between 1 and {len(dd_list)}")
                else:
                    value = dd_list[response_index]['value']
                    ret_dict[name] = value
                    valid_response = True
            else:
                print("Please input a number")

    elif type_hint == 'list':
        valid_response = False
        if isinstance(default, list):
            if not default_is_empty:
                print("\nDefault list")
                for item in default:
                    print(f"  - {item}")
        while not valid_response:
            response = input(f"{desc} (comma seperated list, blank to accept default): ")
            if len(response):
                rl = [x.strip() for x in response.split(',') if len(x)]
            else:
                rl = default
                valid_response = True
                continue
            print("You entered:")
            if len(rl):
                for item in rl:
                    print(f"  - {item}")
            else:
                print("[]")

            if input_yes_no("Are these values correct?", True):
                ret_dict[name] = rl
                valid_response = True

    elif type_hint == 'ip_address':
        valid_response = False
        while not valid_response:
            response = input(f"{var.get('description', name)} {default_str}: ")
            if not len(response):
                response = default
            if is_ip(response):
                valid_response = True
            else:
                print("Invalid IP address.")
        ret_dict[name] = response

    elif type_hint == 'password':
        valid_response = False
        while not valid_response:
            password = getpass(f"{var.get('description', name)}: ")
            confirm = getpass(f"Confirm - {var.get('description', name)}: ")
            if password == confirm:
                valid_response = True
            else:
                print("Password and confirm do not match!")
        ret_dict[name] = password

    elif type_hint == 'text':

        response = input(f"{var.get('description', name)} {default_str}: ")
        if not len(response) and not default_is_empty:
            ret_dict[name] = default
        else:
            ret_dict[name] = response

    elif type_hint == 'hidden':
        ret_dict[name] = default

    elif type_hint == 'fqdn_or_ip':
        response = input(f"{var.get('description', name)} {default_str}: ")
        if not len(response) and not default_is_empty:
            ret_dict[name] = default
        else:
            ret_dict[name] = response

    else:
        raise ValueError(f'Unsupported input variable type of {type_hint} in var {name}')

    return ret_dict


def expandedHomePath(directory):
    """Returns a local OS formatted directory string"""
    return expanduser("~") + os.path.sep + os.path.sep.join(directory.strip().split('/'))



from math import floor
import time

"""
Progress bar that displays percentage complete and tracks time elapsed.
Time elapsed can be disabled by passing clock=False to init.
Best practice is to make sure this is called often, ideally once a second
or more to ensure the clock does not appear to be hung

Usage:
    pb = ProgressBar(prefix="My Task")
    while doing_work():
        do_some_work()
        pb.update(percentage_complete)
    pb.complete()
"""


class ProgressBar:

    def __init__(self, length=50, prefix="Progress", clock=True, fill='█'):
        self.length = length
        self.prefix = prefix
        self.clock = clock
        self.fill = fill
        self.start_time = time.time()
        self.update(0)

    def _get_time(self):
        if not self.clock:
            return ""
        current = floor(time.time() - self.start_time)
        minutes = floor(current / 60)
        seconds = current % 60
        return f"({minutes}m {seconds}s)"

    def update(self, percent):
        """After instantiation, call this to report a new percentage update"""
        filledLength = floor(self.length * (float(percent) / 100))
        bar = self.fill * filledLength + '-' * (self.length - filledLength)
        time_string = self._get_time()
        print(f'\r{self.prefix} |{bar}| {percent}% {time_string}', end="\r")

    def complete(self):
        """Call this when the target workload is complete to display 100%"""
        bar = self.fill * self.length
        time_string = self._get_time()
        print(f'\r{self.prefix} |{bar}| Complete {time_string}\n')



from .base import BaseCommand
from sli.decorators import require_ngfw_connection_params, require_panoply_connection


class DiffCommand(BaseCommand):
    sli_command = "diff"
    short_desc = "Get the differences between two config versions, candidate, running, previous running, etc"
    no_skillet = True
    help_text = """
        Diff module requires 0 to 3 arguments.

        - 0 arguments: Diff running config from previous running config
        - 1 argument: Diff previous config or named config against specified running config
        - 2 arguments: Diff first arg named config against second arg named config
        - 3 arguments: Diff first arg named config against second arg named config and save diffs into the context

        A named config can be either a stored config, candidate, running or a number.
        Positive numbers must be used to specify iterations, 1 means 1 config revision ago

        Example: Get diff between running config and previous running config in set cli format

            user$ sli diff 1 -uc -of set

        Example: Get diff between running config and the candidate config in xml format

            user$ sli diff running candidate -uc -of xml

        Example: Get the diff of all changes made to this device using the autogenerated 'baseline' config, which is
        essentially blank.

            user$ sli diff baseline -uc -of xml

        Example: Get the diff between the second and third most recent running configs

            user$ sli diff 3 2

        Example: Get a diff and save as 'candidate_diff' into the context

            user$ sli diff running candidate candidate_diff -uc
    """

    @require_ngfw_connection_params
    @require_panoply_connection
    def run(self, pan):
        """Get a diff of running and candidate configs"""

        # Any digit arguments should be converted to negative
        def fixup(x):
            return f"-{x}" if x.isdigit() else x

        capture_var = None

        if len(self.args) == 1:
            latest_name = "running"
            source_name = fixup(self.args[0])
        elif len(self.args) == 2:
            source_name = fixup(self.args[0])
            latest_name = fixup(self.args[1])
        elif len(self.args) == 3:
            source_name = fixup(self.args[0])
            latest_name = fixup(self.args[1])
            capture_var = self.args[2]
        elif len(self.args) > 3:
            self._print_usage()
            return
        else:
            # If no arguments supplied, assume diff previous running vs running
            latest_name = "running"
            source_name = "-1"

        previous_config = pan.get_configuration(config_source=source_name)
        latest_config = pan.get_configuration(config_source=latest_name)

        if self.sli.output_format == "xml":
            diff = pan.generate_skillet_from_configs(previous_config, latest_config)
            for obj in diff:
                print(f'name: {obj["name"]}')
                print(f'xpath: {obj["xpath"]}')
                print(f'element: {obj["element"]}\n\n')
        else:
            diff = pan.generate_set_cli_from_configs(previous_config, latest_config)
            for line in diff:
                print(line)

        # Update context if using context
        if self.sli.cm.use_context and capture_var is not None:
            self.sli.context[capture_var] = diff
            print(f'Output added to context as {capture_var}')



import json

import yaml
from jinja2 import Template
from skilletlib import SkilletLoader

from sli.decorators import require_ngfw_connection_params
from .base import BaseCommand

skillet_template = """
name: sli_op
type: pan_validation
snippets:

  - name: op_test
    cmd: {{ cmd }}
    cmd_str: {{ cmd_str }}
    output_type: {{ output_type }}
    outputs:
      - name: op_output
        capture_{{ capture_method }}: {{ capture_arg }}
"""

valid_methods = ["list", "object", "text", "value"]


class OpCommand(BaseCommand):
    sli_command = "op"
    short_desc = "Run an operational command and optionally parse and capture the results into the context"
    no_skillet = True
    help_text = """
    Usage for 'op' module:
        sli op [cmd] [capture_method] [query] [context-variable:optional]

    Example: Get system info and return information as XML text
        sli op "show system info"

    Example: Get system info, using XML command syntax, as a JSON object
        sli op "<show><system><info/></system></show>" object

    Example: get system info as a JSON object and use an XPATH filter for only the plugin_versions
        sli op "show system info" object "./plugin_versions"

    Example: get system info and only return the value of the sw-version tag
        sli op "show system info" value "./sw-version"

    Example: get system info and only return the value of the sw-version tag and store it in the context as 'sw'
        sli op "show system info" text "./sw-version" sw -uc
    """

    @require_ngfw_connection_params
    def run(self):

        # Render validation skillet for execution
        if len(self.args) < 1 or len(self.args) > 4:
            self._print_usage()
            return

        cmd_str: str = self.args[0]
        capture_var: str = ""
        capture_method = "text"
        capture_arg = "."

        if len(self.args) > 1:
            capture_method: str = self.args[1]

        if len(self.args) > 2:
            capture_method: str = self.args[1]
            capture_arg: str = self.args[2]

        if len(self.args) > 3:
            capture_var = self.args[3]

        if capture_method not in valid_methods:
            print(f"Invalid method - {capture_method}")
            self._print_usage()
            return

        if capture_method == "text":
            # either the user said they wanted text output, of they didn't specify
            # if they said text, BUT the specified a query, then what they really want is 'value'
            if capture_arg != ".":
                output_type = "xml"
                capture_method = "value"
            else:
                output_type = "text"
        else:
            output_type = "xml"

        # if the user has passed us an xml string, then use 'op' otherwise use 'cli'
        if cmd_str.startswith("<") and cmd_str.endswith(">"):
            cmd = "op"
        else:
            cmd = "cli"

        skillet_yaml = Template(skillet_template).render(
            {
                "capture_method": capture_method,
                "capture_arg": capture_arg,
                "cmd_str": cmd_str,
                "cmd": cmd,
                "output_type": output_type,
            }
        )

        skillet_dict = yaml.safe_load(skillet_yaml)
        sl = SkilletLoader()
        skillet = sl.create_skillet(skillet_dict)

        # Execute skillet and extract values from target
        exe = skillet.execute(self.sli.context)
        if not skillet.success:
            print("Unable to execute command")
            return

        output = exe["outputs"]["op_output"]
        if capture_method != "text":
            # Print captured JSON
            print(json.dumps(output, indent=4))
        else:
            print(output)

        # Update context if using context
        if self.sli.cm.use_context and len(capture_var) > 1:
            self.sli.context[capture_var] = output
            print(f"Output added to context as {capture_var}")

    def _get_output(self):
        pass



import json

from .base import BaseCommand


class ShowContext(BaseCommand):
    sli_command = 'show_context'
    short_desc = 'Print out contents of an existing context'
    no_skillet = True
    no_context = True

    def run(self):
        context_name = self.sli.options.get('context_name', '')
        if len(context_name) < 1:
            context_name = self.args[0] if len(self.args) > 0 else 'default'
        context = self.sli.cm.load_context(from_file=context_name)
        if not len(context.keys()):
            print(f'Unable to load context {context_name}')
            return
        if self.sli.no_config and "config" in context:
            context.pop('config')
        hidden_fields = ['TARGET_PASSWORD']
        for hf in hidden_fields:
            if hf in context:
                context[hf] = '*****'
        print(json.dumps(context, indent=4))



import paramiko

"""
Provides a raw SSH session with an NGFW, useful for scripting checks against
a user session
"""


class SSHSession:

    BUFFER_LEN = 9999

    def __init__(self, device, username, password, echo=False):
        self.device = device
        self.username = username
        self.password = password
        self.echo = echo
        self._connect()
        self._get_prompt()
        self.mode = "op"
        self.error_text = ""

    def _connect(self):
        self.client = paramiko.client.SSHClient()
        self.client.load_system_host_keys()
        self.client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self.client.connect(
            self.device,
            username=self.username,
            password=self.password
        )
        self.channel = self.client.invoke_shell()

    def _get_prompt(self):
        recvd = ""
        while ">" not in recvd:
            recvd = self.channel.recv(self.BUFFER_LEN).decode()
            if self.echo:
                print(recvd)
        self.prompt = recvd.strip()

    def get_error_text(self):
        """Get error text generated by this module"""
        return self.error_text

    def config_mode(self):
        """Enter configuration mode"""

        self.prompt = self.prompt.replace(">", "#")
        self.channel.send("configure\n")
        recvd = ""
        while self.prompt not in recvd:
            recvd = self.channel.recv(self.BUFFER_LEN).decode()
            if self.echo:
                print(recvd)
        self.mode = "config"

    def set_command(self, cmd):
        """Execute a set command inside config mode, returns False on error"""

        if not self.mode == "config":
            self.config_mode()
        self.channel.send(f"{cmd}\n")
        total_received = ""
        received = ""

        # Scan the output until the prompt is found
        while "[edit]" not in total_received:
            received = self.channel.recv(self.BUFFER_LEN).decode()
            if self.echo:
                print(received)
            total_received += received
            received = ""

        # Check the output for any errors
        error_text = ['Unknown command:', 'Invalid syntax.']
        for et in error_text:
            if et in total_received:
                self.error_text = total_received
                return False
        return True



import click
from sli.skilletLineInterface import SkilletLineInterface
from sli.tools import print_table


class FormatHelp(click.Command):
    def format_help(self, ctx, formatter):
        super().format_help(ctx, formatter)
        commands = SkilletLineInterface.get_commands()
        print('\nAvailable Actions:\n')
        command_list = [{"cmd": x, "desc": getattr(commands[x], "short_desc", None)} for x in commands.keys()]
        print_table(
            command_list,
            {
                "Command": "cmd",
                "Description": "desc"
            })
        print("")

    def parse_args(self, ctx, args):
        """
        Check args for available commands, and print help_text if found. Otherwise,
        default to normal help text via 'format_help'
        """

        if '--help' in args and len(args) > 1:
            commands = SkilletLineInterface.get_commands()
            for command in args:
                if command in commands:
                    c = commands[command]
                    if hasattr(c, 'help_text'):
                        print(c.help_text)
                        ctx.exit()
                    elif hasattr(c, 'short_desc'):
                        print(c.short_desc)
                        ctx.exit()

        super().parse_args(ctx, args)


@click.command(
        cls=FormatHelp,
        context_settings={'allow_extra_args': True},
        no_args_is_help=True
    )
@click.option("-e", "--environment", help="Environment file")
@click.option("-cm", "--commit", is_flag=True, help="Commit configuration changes")
@click.option("-v", "--verbose", is_flag=True, help="Verbose output")
@click.option("-d", "--device", help="Device IP or hostname")
@click.option("-db", "--debug", is_flag=True, help="Run a command in debug mode")
@click.option("-le", "--loader-error", is_flag=True, help="Fail on SkilletLoader errors")
@click.option("-sd", "--directory", help="Directory to load skillets from", default="./")
@click.option("-u", "--username", help="Device username")
@click.option("-p", "--password", help="Device password")
@click.option("-n", "--name", help="Name of skillet to execute")
@click.option("-r", "--report", is_flag=True, help="Generate a panforge formatted report")
@click.option("-rf", "--report-file", help="Location of HTML file to create")
@click.option("-uc", "--use-context", is_flag=True, help="Use a context manager, (global by default)")
@click.option("-cn", "--context-name", help="Use a contexet manager other than global")
@click.option("-ec", "--encrypt-context", is_flag=True, help="Encrypt the context object")
@click.option("-cp", "--context-password", help="Password for encrypted context", default="")
@click.option("-nc", "--no-config", is_flag=True, help="Hide full device configuration from output", )
@click.option("-of", "--output-format", help="Output format, xml or set",
              type=click.Choice(["xml", "set"]), default="xml"
              )
@click.argument("action", nargs=1, default="execute")
@click.pass_context
def cli(ctx, action, **kwargs):

    options = {key: kwargs[key] for key in kwargs if kwargs[key] is not None}
    sli = SkilletLineInterface(options, action, ctx.args)
    sli.execute()


if __name__ == "__main__":
    cli()



import json

import yaml
from jinja2 import Template
from skilletlib import SkilletLoader

from sli.decorators import require_ngfw_connection_params
from .base import BaseCommand

skillet_template = """
name: sli_capture
type: pan_validation
snippets:

  - name: capture_test
    cmd: parse
    variable: config
    outputs:
      - name: capture_test
        capture_{{ capture_method }}: {{ capture_arg }}
"""

valid_methods = ['list', 'object', 'expression']


class Capture(BaseCommand):
    sli_command = 'capture'
    short_desc = 'Capture a value based on object, list, or expression'
    no_skillet = True
    help_text = """
    Usage for capture module:
        sli capture [method] [query] [context-variable:optional]

    valid Methods:
        'list', 'object', 'expression'

    """

    @require_ngfw_connection_params
    def run(self):

        # Render validation skillet for execution
        if len(self.args) < 2 or len(self.args) > 3:
            print(self.help_text)
            return

        capture_method = self.args[0]
        capture_arg = self.args[1]
        capture_var = ''
        if len(self.args) == 3:
            capture_var = self.args[2]
        if capture_method not in valid_methods:
            print(f'Invalid method - {capture_method}')
            print(self.help_text)
            return

        skillet_yaml = Template(skillet_template).render({
            'capture_method': capture_method,
            'capture_arg': capture_arg
        })
        skillet_dict = yaml.safe_load(skillet_yaml)
        sl = SkilletLoader()
        skillet = sl.create_skillet(skillet_dict)
        self.sli.skillet = skillet

        # Execute skillet and extract values from target
        exe = skillet.execute(self.sli.context)
        if not skillet.success:
            print('Unable to execute command')
            return

        output = exe['outputs']['capture_test']

        # Update context if using context
        if self.sli.cm.use_context and len(capture_var) > 1:
            self.sli.context[capture_var] = output
            print(f'Output added to context as {capture_var}')

    def _get_output(self):
        results = self.sli.skillet.get_results()
        # Print captured JSON
        output = results['outputs']
        if 'capture_test' in output:
            print(json.dumps(output['capture_test'], indent=4))



from getpass import getpass
from skilletlib.panoply import Panoply
from skilletlib.exceptions import TargetConnectionException
from sli.tools import get_var
from sli.ssh import SSHSession


def require_ngfw_connection_params(func):
    """
    Validate that SLI command has appropriate contexet paramaters
    to facilitate a connection to an NGFW
    """
    def wrap(command):

        # First, map CLI options into context dict
        device = command.sli.options.get('device', '')
        username = command.sli.options.get('username', '')
        password = command.sli.options.get('password', '')
        if len(device) > 0:
            command.sli.context['TARGET_IP'] = device
        if len(username) > 0:
            command.sli.context['TARGET_USERNAME'] = username
        if len(password) > 0:
            command.sli.context['TARGET_PASSWORD'] = password

        # If we don't have required input parameters, get them from the user
        if len(command.sli.context.get('TARGET_IP', '')) < 1:
            command.sli.context['TARGET_IP'] = input('Device: ')
        if len(command.sli.context.get('TARGET_USERNAME', '')) < 1:
            command.sli.context['TARGET_USERNAME'] = input('Username: ')
        if len(command.sli.context.get('TARGET_PASSWORD', '')) < 1:
            command.sli.context['TARGET_PASSWORD'] = getpass()
        return func(command)
    return wrap


def require_panoply_connection(func):
    """
    Requires a connected panoply session to the target device, will pass in
    the connected panoply instance as an argument
    """
    def wrap(command):
        pan = Panoply(
            command.sli.context['TARGET_IP'],
            command.sli.context['TARGET_USERNAME'],
            command.sli.context['TARGET_PASSWORD']
        )
        if not pan.connected:
            raise TargetConnectionException('Unable to connect to device')
        return func(command, pan)
    return wrap


def require_ngfw_ssh_session(func):
    def wrap(command):
        print(f"Connecting to {command.sli.context['TARGET_IP']}...")
        ssh = SSHSession(
            command.sli.context['TARGET_IP'],
            username=command.sli.context['TARGET_USERNAME'],
            password=command.sli.context['TARGET_PASSWORD']
        )
        print("Connected.")
        return func(command, ssh)
    return wrap


def load_variables(func):
    """Load variables from skillet and get user input if not supplied"""

    def wrap(command):
        for var in command.sli.skillet.variables:
            get_var(var, command.args, command.sli.context)
        if len(command.sli.skillet.variables):
            print("End of user variables.")
        return func(command)
    return wrap


def require_single_skillet(func):
    """Commands decorated with this require one skillet to be uniquely specified"""

    def wrap(command):
        if not command.sli.options.get('name') and len(command.sli.skillets) > 1:
            print(f'Specify a skillet to run with --name or -n when more than 1 is present for command {command.sli_command}')
            exit(1)

        target_name = command.sli.options.get('name') if command.sli.options.get('name') else command.sli.skillets[0].name
        command.sli.skillet = command.sli.sl.get_skillet_with_name(target_name)
        if command.sli.skillet is None:
            print(f'Unable to load skillet {target_name} by name')
            exit(1)
        return func(command)
    return wrap


def require_skillet_type(*args):
    """
    Commands decorated with this require a specific type of skillet to be executed.
    Inputs expected as *args, selected skillet type must match one
    """
    def inner(func):
        def wrap(command):
            skillet_type = command.sli.skillet.type
            if skillet_type not in args:
                print(f'Invalid type of skillet ({skillet_type}) for command {command.sli_command}, requires one of: {", ".join(args)}')
                exit(1)
            return func(command)
        return wrap
    return inner



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="sli",
    version="0.0.0.0a5",
    author="Andrew Mallory",
    author_email="amallory@paloaltonetworks.com",
    description="Document generation library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://paloaltonetworks.com",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
    install_requires=[
        'skilletlib',
        'panforge',
        'click',
        'pycryptodome',
        'paramiko',
    ],
    entry_points={
        'console_scripts': [
            'sli= sli.cli:cli',
        ]
    },
)



from .base import BaseCommand


class ClearContext(BaseCommand):
    """
    Clear SLI context.
    Pass the 'creds' argument to only remove non credential parameters
    """

    sli_command = 'clear_context'
    short_desc = 'Clear out contents of a specific context'
    no_skillet = True
    no_context = True

    def run(self):

        # Find the name of the target context
        args = [x for x in self.args if not x == 'creds']
        context_name = self.sli.options.get('context_name', '')
        if len(context_name) < 1:
            context_name = args[0] if len(args) > 0 else 'default'

        # Remove non credential items from existing context
        if 'creds' in self.args:
            if self.sli.cm.clean_context(context_name):
                print(f"Non credential items have been removed from context {context_name}")

        # Remove the entire context by default
        else:
            if self.sli.cm.remove_context(context_name):
                print(f'Context {context_name} has been removed')



import os

from sli.decorators import load_variables
from sli.decorators import require_ngfw_connection_params
from sli.decorators import require_single_skillet
from sli.decorators import require_skillet_type
from sli.tools import generate_report
from sli.tools import print_table
from .base import BaseCommand

"""
Execute a validation skillet and generate results
"""


class ValidateCommand(BaseCommand):
    sli_command = 'validate'
    short_desc = 'Execute a validation skillet of type pan_validation'

    @require_single_skillet
    @require_skillet_type('pan_validation')
    @require_ngfw_connection_params
    @load_variables
    def run(self):
        """
        SLI action, test load all skillets in directory and print out loaded skillets
        Generate panforge report if requested
        """

        print(f"Executing {self.sli.skillet.name}...")
        self.sli.skillet.execute(self.sli.context)

    def _get_output(self):
        exe = self.sli.skillet.get_results()
        if self.sli.verbose:
            self.print_verbose(exe)
        self.print_summary(exe)

        if getattr(self.sli, 'generate_report', False):
            header = {'Host': self.sli.context['TARGET_IP']}
            out_file = getattr(self.sli, 'report_file', '')
            if len(out_file) < 1:
                out_file = f'{self.sli.options["TARGET_IP"]}-{self.sli.skillet.name}.html'
            report_dir = os.path.sep.join([self.sli.skillet.path, 'report'])
            generate_report(out_file, exe['pan_validation'], report_dir, header=header)

    @staticmethod
    def print_verbose(exe):
        print('Validation details\n------------------')
        for snippet_name in exe['pan_validation']:
            snippet = exe['pan_validation'][snippet_name]
            print(snippet_name)
            print('-' * len(snippet_name))
            if 'results' in snippet:
                result = "Passed" if snippet['results'] else "Failed"
                print(f"   Validation Results: {result}")
            print(f"   Label: {snippet['label']}")
            print(f"   Output: {snippet['output_message']}")
            documentation_link = snippet.get('documentation_link')
            if documentation_link:
                print(f"   Documentation Link: {documentation_link}")
            if isinstance(snippet['meta'], dict):
                print('   Meta:')
                for key in snippet['meta']:
                    print(f"      {key}: {snippet['meta'][key]}")
            print()

    @staticmethod
    def print_summary(exe):
        results = [
            {
                'name': x,
                'result': "Passed" if exe['snippets'][x] else "Failed"
            }
            for x in exe['snippets']
        ]
        print('Validation results\n------------------')
        print_table(results, {
            "Name": "name",
            "Result": "result"
        })
